library(XML)
library(RCurl)
library(reshape)
library(reshape2)
library(dplyr)
library(tidyr)
options(stringsAsFactors=FALSE)


#==========================================================
# Auxiliary functions
#==========================================================

doit <- function(x){
  url <- "StandardGraphsWebServices.asmx/getSummaryTable?key="
	url <- paste0("http://standardgraphs.ices.dk/", url, x$key[1])
	obj <- getURL(url)
	obj <- xmlToList(xmlTreeParse(obj)$doc$children[[1]])
	df0 <- as.data.frame(apply(obj$lines, 1, as.numeric))
	df0 <- cbind(df0, do.call("cbind", obj[names(obj)[!names(obj) == "lines"]]))
  url <- "StandardGraphsWebServices.asmx/getFishStockReferencePoints?key="
	url <- paste0("http://standardgraphs.ices.dk/", url, x$key)
	obj <- getURL(url)
	df0 <- cbind(df0, xmlToDataFrame(obj))
	df0 <- melt(df0, id.vars="Year")
	df0 <- cbind(df0, x)
	return(df0)	
}

doitAreas <- function(x){
	url <- paste("http://vocab.ices.dk/services/pox/GetCodeDetail/ices_stockcode/",
    x$Key[1], sep="")
	ox0 <- getURL(url)
	ox0 <- xmlTreeParse(ox0)
	ox0 <- xmlToList(ox0$doc$children$GetCodeDetailResponse$children$CodeDetail)
	s0 <- data.frame(t(unlist(ox0[1:5])))
	a0 <- do.call("rbind", lapply(lapply(ox0[7:length(ox0)], "[", "Code"),
    unlist))
	a1 <- do.call("rbind", lapply(lapply(ox0[7:length(ox0)], "[", "CodeType"),
    unlist))
	a0 <- a0[a1[,"CodeType.Key"]=="ICES_Area",]
	p0 <- data.frame(t(unlist(ox0[[6]]$Code)))
	s0 <- data.frame(s0, p0)
	if(is.null(nrow(a0))){
		ox <- data.frame(s0, t(a0))
	} else {
		s0 <- s0[rep(1, nrow(a0)),]
		ox <- cbind(s0, a0)
	}
	ox <- ox[,c("Key", "Description", "Description.1", "Code.Key",
    "Code.Description")]
	names(ox) <- c("stock", "stockDescription", "species", "area",
    "areaDescription")
	return(ox)
}

###########################################################
# Mediterranean
###########################################################
#==========================================================
# Mediterranean stock assessment information
#==========================================================
mmu <- read.csv("../data/medManagementUnits.csv")
mmu$manUnit <- paste(mmu$Stock, mmu$Area, sep="_")
mmu <- subset(mmu, !(Meeting %in% c("10-05_SG-MED 10-02", "2011-11_STECF 11-14")))

#==========================================================
# Mediterranean management units (MLS) information
#==========================================================
load("../data/Mediterranean_cfp2015.Rdata")
msa <- cfp2015
msa$stk <- paste(msa$Stock, msa$Area, sep="_")
msa$StockDescription <- paste(msa$Species, "in GSA", gsub("_", ", ",  msa$Area))

###########################################################
# NE Atlantic
###########################################################
#==========================================================
# ICES stock assessment information
#==========================================================
#----------------------------------------------------------
# get stocks list
#----------------------------------------------------------
url <- "StandardGraphsWebServices.asmx/getListStocks?year=0"
stks <- getURL(paste0("http://standardgraphs.ices.dk/", url))
stks <- xmlToDataFrame(stks)
# Remove unpublished,
stks <- subset(stks, Status!="Not Published")
# "Psetta maxima (historic name) North Sea"
stks <- subset(stks, SpeciesName!="Psetta maxima (historic name)")
# and Molva molva in Iceland
stks <- subset(stks, !(SpeciesName=="Molva molva" &
  EcoRegion=="Iceland and East Greenland"))
# and had-34 and had-scow, which were replaced by had-346a but were left in the db
stks <- subset(stks, !(FishStockName %in% c("had-34", "had-scow")))
# get most recent assessment
stks <- lapply(split(stks, as.character(stks$FishStockName)), 
	function(x){
		subset(x, x$AssessmentYear==max(as.numeric(as.character(x$AssessmentYear))))
	})

#----------------------------------------------------------
# get stock summaries and reference points
#----------------------------------------------------------
stks <- lapply(stks, function(x) try(doit(x)))
stks <- do.call("rbind", stks[!unlist(lapply(stks, is , "try-error"))])
vars <- c("SSB", "SSB.nil", "landings", "F.nil", "F", "catches.nil", "catches",
  "discards.nil", "discards")
cond <- c("Fage", "FLim", "Fpa", "Bpa", "Blim", "FMSY", "MSYBtrigger", "units",
  "stockSizeDescription", "stockSizeUnits", "fishingPressureDescription",
  "fishingPressureUnits", "MSYBescapement", "Fmanagement", "Bmanagement")

# object
isa <- dcast(subset(stks, variable %in% c(vars, cond)), ... ~variable)
# set empty to NA
isa[isa==""] <- NA

# merge columns with ".nil" into the correct place
isa[is.na(isa$SSB),"SSB"] <- isa[is.na(isa$SSB),"SSB.nil"]
isa[is.na(isa$F),"F"] <- isa[is.na(isa$F),"F.nil"]
isa[is.na(isa$catches),"catches"] <- isa[is.na(isa$catches),"catches.nil"]
isa[is.na(isa$discards),"discards"] <- isa[is.na(isa$discards),"discards.nil"]
isa$SSB.nil <- isa$F.nil <- isa$catches.nil <- isa$discards.nil <- NULL
	
# fix codes for fishingPressureDescription
isa[!(isa$fishingPressureDescription %in% c("Fishing Pressure: F/FMSY",
  "Fishing Pressure: F", "Fishing pressure: Relative")),"fishingPressureDescription"] <-
  "Fishing Pressure: Harvest rate"

# fix codes for EcoRegion
isa[isa$EcoRegion=="North Sea", "EcoRegion"] <- "Greater North Sea"

# coerce some numerical variables
isa$SSB <- as.numeric(isa$SSB)
isa$F <- as.numeric(isa$F)
isa$FMSY <- as.numeric(isa$FMSY)
isa$Bpa <- as.numeric(isa$Bpa)
isa$Blim <- as.numeric(isa$Blim)
isa$MSYBtrigger <- as.numeric(isa$MSYBtrigger)
isa$MSYBescapement <- as.numeric(isa$MSYBescapement)

# consistency for FMSY when F reported has ratio
isa[isa$fishingPressureDescription == "Fishing Pressure: F/FMSY","FMSY"] <- 1
# drop "Fishing pressure: Relative" (relative to the mean, ICES dixit)
isa[isa$fishingPressureDescription == "Fishing pressure: Relative","FMSY"] <- NA

#==========================================================
# ICES Stock assessment units definition
#==========================================================
#----------------------------------------------------------
# get data
#----------------------------------------------------------
stkAreas <- xmlToDataFrame(
  getURL("http://vocab.ices.dk/services/pox/GetCodeList/ICES_STockCode"))
stkAreas <- lapply(split(stkAreas, stkAreas$Key), function(x) try(doitAreas(x)))
stkAreas <- do.call("rbind", stkAreas)

#----------------------------------------------------------
# parse codes
#----------------------------------------------------------

# Subarea and Division
stkAreas$FAOArea <- 27
stkAreas$Subarea <- NA
stkAreas$Division <- NA
stkAreas$Subdivision <- NA

# Subarea codes in right order for parsing
sas <- c("XIV", "XIII", "XII", "XI", "IX", "X", "VIII", "VII", "VI", "IV", "V",
  "III", "II", "I")

rows <- seq(dim(stkAreas)[1])

for(i in sas) {
  # rows matching Subarea i
  idx <- grep(i, stkAreas$area[rows], fixed=TRUE)
  stkAreas[rows,]$Subarea[idx] <- i
  lst <- strsplit(sub(i, "", stkAreas[rows,]$area[idx]), "")
  stkAreas[rows,]$Division[idx] <- unlist(lapply(lst, "[", 1))
  stkAreas[rows,]$Subdivision[idx] <- unlist(lapply(lst, "[", 2))
  rows <- rows[!rows %in% rows[idx]]
}

# Baltic
idx <- !is.na(as.numeric(as.character(stkAreas$area)))
stkAreas[idx,"Subarea"] <- "III"
stkAreas[idx,"Division"] <- as.character(stkAreas[idx,"area"])
idx <- stkAreas$area %in% c("28-1","28-2")
stkAreas[idx,"Subarea"] <- "III"
stkAreas[idx,"Division"] <- 28
stkAreas[idx,"Subdivision"] <- sub("28-", "", stkAreas[idx,"area"]) 

#==========================================================
# TAC units
#==========================================================
#----------------------------------------------------------
# merge areas and species definitions into TAC Quotas
#----------------------------------------------------------

# TAC Quotas
tac <- read.csv("../data/TacQuotas20160125095252.csv", sep=";", header=TRUE)

# Areas
area <- read.csv("../data/STOCKS_v3_4.tab", sep="\t", header=TRUE,
  stringsAsFactors=FALSE)
area$Area <- area$STOCK_ID
area$nc  <- nchar(area$STOCK_ID)
area$Area <- substr(area$Area, 5, area$nc)
area$Description <- area$Stock_Area
area <- subset(area, Year_ID==2014)
area <- unique(area[,c("Area","Description")])
area <- do.call("rbind", lapply(split(area, area$Area), "[", i=1, j=1:2))

# Bring area into tac
tac <- merge(tac,area[,c("Area","Description")])

# Species
species <- read.csv("../data/species_headers_20160125.xls", sep="\t", header=TRUE)
colnames(species)[colnames(species) == "Description"] <- "species_desc"

tac <- merge(tac, species[,c("Species", "species_desc", "Latin.name")])

# We want tac$Stock.Group to be T-Baltic, T-WWN and T-DSS
tacsub <- tac[tac$Stock.Group %in% c("T-BALTIC", "T-WWN", "T-DSS"),]

#----------------------------------------------------------
# The breakdown of the management unit areas into FAO/ICES 
# areas had to be done manually. The source data was in 
# "tacDefs.csv" and the results in 
# "TacQuotas20160125095252-AreasSelected_WWN_Baltic.csv".
#----------------------------------------------------------
# INPUT: CSV file of TAC areas
# Empty strings as NA

inp <- read.csv('../data/TacQuotas20160222-AreasSelected_WWN_Baltic_DSS.csv',
  stringsAsFactors=FALSE, na.strings="", sep=';') 
# Drop NAs in VAR, SAD (all NAs in ICES1-18)
itac <-  gather(inp, "VAR", "SAD", 13:30, na.rm=TRUE)
itac <- select(itac, Species, Area, Latin.name, SAD)

# HACK: Delete "NA" from Latin.name
itac$Latin.name[itac$Latin.name == "NA"] <- ""

# Subarea and Division
itac$FAOArea <- 27
itac$Subarea <- ""
itac$Division <- ""

# Subarea codes in right order for parsing
sas <- c("XIV", "XIII", "XII", "XI", "IX", "X", "VIII", "VII", "VI", "IV", "V",
  "III", "II", "I")
rows <- seq(dim(itac)[1])

for(i in sas) {
  # rows matching Subarea i
  idx <- grep(i, itac$SAD[rows], fixed=TRUE)
  itac[rows,]$Subarea[idx] <- i
  itac[rows,]$Division[idx] <- sub(i, "", itac[rows,]$SAD[idx])
  rows <- rows[!rows %in% rows[idx]]
}

# Divisions by Subarea
divs <- list(I=2, II=2, III=4, IV=3, V=2, VI=2, VII=10, VIII=5, IX=2, X=2,
  XI=0, XII=3, XIII=0, XIV=2)
divs <- lapply(divs, function(x) letters[-9][seq(length=x)])

# BUT ICES 22-32 for Baltic (III)
divs[['III']] <- as.character(seq(22,32))

# EXPAND all subarea divisions
idx <- itac$Division == ""
sub <- itac[idx,]
res <- vector('list', length=nrow(sub))

for (i in seq(nrow(sub))) {
  ds <- divs[[sub[i,'Subarea']]]
  res[[i]] <- suppressWarnings(cbind(sub[i, -7], Division=ds))
}
res <- Reduce('rbind', res)

# JOIN res and !sub
itac <- rbind(res, itac[!idx,])

# asterisk? special = T/F
itac$special <- grepl('*', itac$Area, fixed=TRUE)

# fix name
itac[itac$Latin.name=="Merlangius merlangus, Pollachius pollachius" &
  !is.na(itac$Latin.name),"Latin.name"] <- "Merlangius merlangus"

#==========================================================
# Merging ICES and EU Mediterranean datasets
#==========================================================
#----------------------------------------------------------
# Stock assessment summaries
#----------------------------------------------------------
sa <- isa

# fix ecoregion
sa[sa$EcoRegion %in% c("Bay of Biscay and Iberian Sea",
  "Celtic Sea and West of Scotland"), "EcoRegion"] <- "Western European"
sa[sa$EcoRegion=="Widely distributed and migratory stocks",
  "EcoRegion"] <- "Widely distributed"
sa[sa$EcoRegion %in% c("Iceland and East Greenland",
  "Barents Sea and Norwegian Sea", "Faroe Plateau Ecosystem"),
  "EcoRegion"] <- "Other"
# stocks in the western channel are wrongly classified in 
# western waters, must be Greater North Sea
sa[grep("echw", sa$FishStockName),"EcoRegion"] <- "Greater North Sea"

# add Med
df0 <- sa[1:nrow(msa),]
df0[] <- NA
df0[,c("Year", "FishStockName", "StockDescription", "SSB", "F", "AssessmentYear", "FMSY",
  "SpeciesName", "landings")] <- msa[,c("year", "stk", "StockDescription", "SSB", "F",
  "asses_year", "Fref", "Scientific_name", "Landings")]
df0$EcoRegion <- "Mediterranean"
# all stocks in the list have analyctical assessments
df0$fishingPressureDescription <- "Fishing Pressure: F"
sa <- rbind(sa, df0)

###########################################################
# Filtering and processing data
###########################################################
#==========================================================
# Bref
#==========================================================
# Blim << MSYBtrigger << Bpa as Bref
sa$Bref <- sa$Blim
sa$Bref[!is.na(sa$MSYBtrigger)] <- sa$MSYBtrigger[!is.na(sa$MSYBtrigger)]
sa$Bref[!is.na(sa$Bpa)] <- sa$Bpa[!is.na(sa$Bpa)]
# B escapment as Bref for relevant stocks
sa$Bref[!is.na(sa$MSYBescapement)] <- sa$MSYBescapement[!is.na(sa$MSYBescapement)]
sa$Bref <- as.numeric(sa$Bref)
# set 0 as NA
sa$Bref[sa$Bref==0] <- NA

#==========================================================
# Fref
#==========================================================
sa$Fref <- sa$FMSY
# no Fref f B escapment 
sa$Fref[!is.na(sa$MSYBescapement)] <- NA
sa$Fref <- as.numeric(sa$Fref)
sa$Fref[sa$Fref==0] <- NA

#==========================================================
# COMPUTE F/Fref and B/Bref | year + stock
#==========================================================
sa <- transform(sa, indF = F/Fref, indB=SSB/Bref)
# in case of escapement strategy MSY evaluated by SSB ~ Blim/Bpa/etc
sa$indF[!is.na(sa$MSYBescapement)] <- sa$Bref[!is.na(sa$MSYBescapement)] /
  sa$SSB[!is.na(sa$MSYBescapement)]
sa <- transform(sa, sfFind=!is.na(indF))

#==========================================================
# COMPUTE SBL | year + FishStock
#==========================================================
sa$SBL <- !(sa$indF > 1 | sa$indB < 1)
# if one is NA SBL can't be inferred
sa$SBL[is.na(sa$indF) | is.na(sa$indB)] <- NA

#==========================================================
# Mapping management units (TAC or MLS) with stock 
# assessment units
#==========================================================
#----------------------------------------------------------
# ICES info
#----------------------------------------------------------
isu <- subset(stkAreas, stock %in% unique(sa[sa$sfFind,]$FishStockName))
# create genus field
isu$genus <- unlist(lapply(strsplit(as.character(isu$species), " "), "[", 1))

#----------------------------------------------------------
# TAC info
#----------------------------------------------------------
# remove special conditions and species NA
imu <- itac[!itac$special,]
imu <- imu[!is.na(imu$Latin.name),]
# some genus need correction
imu[imu$Latin.name == "Caproidae", "Latin.name"] <- "Capros"
imu[imu$Latin.name == "Lophiidae", "Latin.name"] <- "Lophius"
# create genus field
imu$genus <- unlist(lapply(strsplit(imu$Latin.name, " "), "[", 1))
# create tacUnit field
imu <- transform(imu, tacUnit = paste(Species, Area, sep=":"))
# remove TACs that are not set by the EU
v0 <- c("COD:2A3AX4", "COD:5W6-14", "COD:GRL1", "COD:GRL2", "COD:GRL3", "COD:GRL4",
  "HAD:2AC4.", "HAD:5BC6A.", "USK:1214EI", "USK:1214EI", "USK:1214EI", "USK:1214EI",
  "USK:1214EI", "USK:1214EI")
imu <- subset(imu, !(tacUnit %in% v0))

#----------------------------------------------------------
# merge
#----------------------------------------------------------
imu <- lapply(split(imu, imu$Latin.name), function(x){
	spp <- tolower(sub(" ", "", x$Latin.name[1]))
	aa <- unique(isu[tolower(sub(" ", "", isu$species)) == spp,
    c("stock","Subarea", "Division")])
	# need to deal with more than one stock assessment for a single tac unit
	aa <- cast(aa, Subarea + Division ~ ., paste, collapse=";", value="stock")
	aa <- as.data.frame(aa)
	names(aa) <- c("Subarea", "Division", "stock")	
	# combine
	x <- merge(x,aa, all.x=TRUE)
	x[,"subareaAssessed"] <- FALSE
	x[,"genusAssessed"] <- FALSE
	x[x$Subarea %in% aa$Subarea,"subareaAssessed"] <- TRUE
	xx <- x[!x$subareaAssessed,]
	if(nrow(xx)> 1 & !all(xx$Latin.name %in% c("Pollachius pollachius", "Pollachius virens"))){
		xx$stock <- NULL
		gen <- tolower(xx$genus[1])
		aa <- unique(isu[tolower(isu$genus) == gen,c("stock","Subarea", "Division")])
		# need to deal with more than one stock assessment for a single tac unit
		aa <- cast(aa, Subarea + Division ~ ., paste, collapse=";", value="stock")
		aa <- as.data.frame(aa)
		names(aa) <- c("Subarea", "Division", "stock")	
		xx <- merge(xx, aa, all.x=TRUE)
		xx[xx$Subarea %in% aa$Subarea,"subareaAssessed"] <- TRUE
		# in these cases the TAC unit combines spp so assessment is at genus level
		xx[xx$Subarea %in% aa$Subarea,"genusAssessed"] <- TRUE
		x <- rbind(x[x$subareaAssessed,],xx)
	}
	x[,"divisionAssessed"] <- !is.na(x$stock)
	x
})

imu <- do.call("rbind", imu)

#==========================================================
# Merging ICES and EU Mediterranean datasets
#==========================================================
#----------------------------------------------------------
# Stock assessment coverage
#----------------------------------------------------------
# add EcoRegion
mu <- imu[,c("Latin.name", "genus", "Species", "FAOArea", "Subarea",
  "Division", "Area", "tacUnit", "subareaAssessed", "genusAssessed",
  "divisionAssessed", "stock")]
mu[, "EcoRegion"] <- "Other"
mu[mu$Subarea=="III", "EcoRegion"] <- "Baltic Sea"
mu[mu$Subarea=="IV" | (mu$Subarea=="III" & mu$Division=="a") | (mu$Subarea=="VII" &
  mu$Division=="d") | (mu$Subarea=="VII" &
  mu$Division=="e"), "EcoRegion"] <- "Greater North Sea"
mu[mu$Subarea %in% c("VI", "VII", "VIII", "IX"),
  "EcoRegion"] <- "Western European"
mu[mu$stock %in% unique(subset(isa,
  EcoRegion =="Widely distributed and migratory stocks")$FishStockName),
  "EcoRegion"] <- "Widely distributed"

df0 <- mu[1:nrow(mmu),]
df0$FAOArea <- 37
df0$EcoRegion <- "Mediterranean"
df0$genus <- df0$Division <- df0$divisionAssessed <- NA
df0[,c("Latin.name", "Species", "Subarea", "tacUnit", "subareaAssessed",
  "genusAssessed", "stock")] <- mmu[,c("Scientific_name", "Stock", "Area",
  "manUnit", "Assessed", "genusAssessed", "Stock_Assessed_GSAs")]
# note that in the Med the management area matches the FAO subarea (GSA)
df0$Area <- df0$Subarea
mu <- rbind(mu, df0)
# fix some names to be more explicit and/or short
names(mu) <- c("spp.latin", "genus", "spp.fao", "FAOArea", "subarea",
  "division", "manArea", "manUnit", "subareaAssessed", "genusAssessed",
  "divisionAssessed", "stockUnit", "EcoRegion")

#==========================================================
# Clean and save
#==========================================================

save(mu, sa, file="data.RData")

